// SPDX-License-Identifier: MIT
// SPDX-FileCopyrightText: 2025 Loïc Le Page <llepage@igalia.com>

#include "Camera.h"

std::shared_ptr<Camera> Camera::create(unsigned int captureWidth, unsigned int captureHeight) noexcept
{
    // Initialize the camera manager.
    libcamera::logSetTarget(libcamera::LoggingTargetNone);

    std::shared_ptr<Camera> camera(new Camera());
    camera->m_cameraManager = std::make_unique<libcamera::CameraManager>();
    camera->m_cameraManager->start();

    // Select the first available camera.
    auto cameras = camera->m_cameraManager->cameras();
    if (cameras.empty())
    {
        std::cerr << "No camera found" << std::endl;
        return nullptr;
    }
    camera->m_selectedCamera = cameras[0];

    // Lock the selected camera and choose a configuration for video display.
    camera->m_selectedCamera->acquire();

    auto camConfig = camera->m_selectedCamera->generateConfiguration({libcamera::StreamRole::Viewfinder});
    if (camConfig->empty())
    {
        std::cerr << "No suitable configuration found for the selected camera" << std::endl;
        return nullptr;
    }

    auto& streamConfig = camConfig->at(0);
    std::cout << "Default camera configuration is: " << streamConfig.toString() << std::endl;

    // We need to choose a pixel format with a stride aligned on 32 bits to be
    // compatible with the GLES renderer. We only need 2 buffers, while one
    // buffer is used by the GLES renderer, the other one is filled by the
    // camera next frame and then both buffers are swapped.
    streamConfig.size.width = captureWidth;
    streamConfig.size.height = captureHeight;
    streamConfig.pixelFormat = libcamera::formats::XRGB8888;
    streamConfig.bufferCount = 2;

    if (camConfig->validate() == libcamera::CameraConfiguration::Invalid)
    {
        std::cerr << "Invalid camera configuration" << std::endl;
        return nullptr;
    }
    std::cout << "Targeted camera configuration is: " << streamConfig.toString() << std::endl;

    if (camera->m_selectedCamera->configure(camConfig.get()) != 0)
    {
        std::cerr << "Failed to update the camera configuration" << std::endl;
        return nullptr;
    }
    std::cout << "Camera configured successfully" << std::endl;

    // Allocate the buffers pool used to fetch frames from the camera.
    libcamera::Stream* stream = streamConfig.stream();
    camera->m_frameAllocator = std::make_unique<libcamera::FrameBufferAllocator>(camera->m_selectedCamera);
    if (camera->m_frameAllocator->allocate(stream) < 0)
    {
        std::cerr << "Failed to allocate buffers for the selected camera stream" << std::endl;
        return nullptr;
    }

    auto& buffersPool = camera->m_frameAllocator->buffers(stream);
    std::cout << "Camera stream has a pool of " << buffersPool.size() << " buffers" << std::endl;

    // Create the requests used to fetch the actual camera frames.
    for (auto& buffer : buffersPool)
    {
        auto request = camera->m_selectedCamera->createRequest();
        if (!request)
        {
            std::cerr << "Failed to create a frame request for the selected camera" << std::endl;
            return nullptr;
        }

        if (request->addBuffer(stream, buffer.get()) != 0)
        {
            std::cerr << "Failed to add a buffer to the frame request" << std::endl;
            return nullptr;
        }

        camera->m_requests.push_back(std::move(request));
    }

    // Connect the requests execution callback, it is called each time a frame
    // has been produced by the camera.
    camera->m_selectedCamera->requestCompleted.connect(camera.get(), &Camera::onRequestCompleted);

    return camera;
}

Camera::~Camera()
{
    if (m_selectedCamera)
    {
        m_selectedCamera->stop();
        m_selectedCamera->requestCompleted.disconnect(this);
        m_selectedCamera->release();
    }

    {
        std::lock_guard<std::mutex> lock(m_nextFrameMutex);
        m_nextFrame.reset();
    }

    if (m_cameraManager)
    {
        m_cameraManager->stop();
    }
}

void Camera::startCapturing() const noexcept
{
    m_selectedCamera->start();
    for (const auto& request : m_requests)
    {
        request->reuse(libcamera::Request::ReuseBuffers);
        m_selectedCamera->queueRequest(request.get());
    }
}

void Camera::stopCapturing() const noexcept
{
    m_selectedCamera->stop();
}

Camera::Frame::~Frame()
{
    auto camera = m_camera.lock();
    if (camera && m_request)
    {
        m_request->reuse(libcamera::Request::ReuseBuffers);
        camera->m_selectedCamera->queueRequest(m_request);
    }
}

std::unique_ptr<Camera::Frame> Camera::getNextFrame() noexcept
{
    std::lock_guard<std::mutex> lock(m_nextFrameMutex);
    return std::move(m_nextFrame);
}

void Camera::onRequestCompleted(libcamera::Request* request)
{
    if (request->status() == libcamera::Request::RequestCancelled)
    {
        return;
    }

    // We can directly take the first request buffer as we are managing
    // only one stream. In case of multiple streams, we should iterate
    // over the BufferMap entries or access the buffer by stream pointer.
    auto buffer = request->buffers().begin()->second;
    if (buffer->metadata().status == libcamera::FrameMetadata::FrameSuccess)
    {
        // As we are using a XRGB8888 color format we have only one plane, but
        // in case of using a multiplanes color format (like YUV420) we
        // should iterate over all the planes.
        std::unique_ptr<Frame> frame(new Frame(this, request, buffer->cookie()));

        std::lock_guard<std::mutex> lock(m_nextFrameMutex);
        m_nextFrame = std::move(frame);
    }
    else
    {
        // Reuse the request buffer and re-queue the request.
        request->reuse(libcamera::Request::ReuseBuffers);
        m_selectedCamera->queueRequest(request);
    }
}
