// SPDX-License-Identifier: MIT
// SPDX-FileCopyrightText: 2025 Loïc Le Page <llepage@igalia.com>

#include "CameraTexture.h"

#include <EGL/eglext.h>
#include <GLES2/gl2ext.h>

static PFNGLEGLIMAGETARGETTEXTURE2DOESPROC glEGLImageTargetTexture2DOES = nullptr;

std::unique_ptr<CameraTexture> CameraTexture::create(EGLDisplay eglDisplay,
                                                     const libcamera::StreamConfiguration& streamConfiguration,
                                                     const libcamera::FrameBuffer& buffer) noexcept
{
    // Initialize the OES_EGL_image_external extension.
    if (!glEGLImageTargetTexture2DOES)
    {
        glEGLImageTargetTexture2DOES =
            reinterpret_cast<PFNGLEGLIMAGETARGETTEXTURE2DOESPROC>(eglGetProcAddress("glEGLImageTargetTexture2DOES"));
        if (!glEGLImageTargetTexture2DOES)
        {
            return nullptr;
        }
    }

    // Create an EGLImage from the camera FrameBuffer.
    // In our case we are using a packed color format (XRGB8888), so we
    // only need the first buffer plane. In case of using a multiplanar color
    // format (like YUV420 for example), we would need to iterate over all the
    // color planes in the buffer and fill the EGL_DMA_BUF_PLANE[i]_FD_EXT,
    // EGL_DMA_BUF_PLANE[i]_OFFSET_EXT and EGL_DMA_BUF_PLANE[i]_PITCH_EXT for
    // each plane.
    const auto& plane = buffer.planes().front();

    const EGLAttrib attrs[] = {EGL_WIDTH,
                               streamConfiguration.size.width,
                               EGL_HEIGHT,
                               streamConfiguration.size.height,
                               EGL_LINUX_DRM_FOURCC_EXT,
                               streamConfiguration.pixelFormat.fourcc(),
                               EGL_DMA_BUF_PLANE0_FD_EXT,
                               plane.fd.get(),
                               EGL_DMA_BUF_PLANE0_OFFSET_EXT,
                               (plane.offset != libcamera::FrameBuffer::Plane::kInvalidOffset) ? plane.offset : 0,
                               EGL_DMA_BUF_PLANE0_PITCH_EXT,
                               streamConfiguration.stride,
                               EGL_NONE};

    EGLImage eglImage = eglCreateImage(eglDisplay, EGL_NO_CONTEXT, EGL_LINUX_DMA_BUF_EXT, nullptr, attrs);
    if (!eglImage)
    {
        return nullptr;
    }

    std::unique_ptr<CameraTexture> texture(new CameraTexture());

    // Create the GLES texture and attach the EGLImage to it.
    glGenTextures(1, &texture->m_texture);
    glBindTexture(GL_TEXTURE_EXTERNAL_OES, texture->m_texture);
    glTexParameteri(GL_TEXTURE_EXTERNAL_OES, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_EXTERNAL_OES, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_EXTERNAL_OES, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_EXTERNAL_OES, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    glEGLImageTargetTexture2DOES(GL_TEXTURE_EXTERNAL_OES, eglImage);
    glBindTexture(GL_TEXTURE_EXTERNAL_OES, 0);

    // Now that the EGLImage is attached to the texture, we can destroy it. The
    // underlying dma-buf will be released when the texture is deleted.
    eglDestroyImage(eglDisplay, eglImage);

    return texture;
}

CameraTexture::~CameraTexture()
{
    if (m_texture)
    {
        glDeleteTextures(1, &m_texture);
    }
}

void CameraTexture::bind() const noexcept
{
    glBindTexture(GL_TEXTURE_EXTERNAL_OES, m_texture);
}
