// SPDX-License-Identifier: MIT
// SPDX-FileCopyrightText: 2025 Loïc Le Page <llepage@igalia.com>

#include "Camera.h"
#include "CameraTexture.h"
#include "rendering/AppWindow.h"
#include "rendering/CubeMesh.h"
#include "rendering/PhongShader.h"

#include <glm/gtc/matrix_transform.hpp>

class AppRenderer final : public IRenderer
{
  public:
    AppRenderer(unsigned int width, unsigned int height) : m_width(width), m_height(height)
    {
    }

    ~AppRenderer() override
    {
        onShut();
    }

    AppRenderer(AppRenderer&&) = delete;
    AppRenderer& operator=(AppRenderer&&) = delete;
    AppRenderer(const AppRenderer&) = delete;
    AppRenderer& operator=(const AppRenderer&) = delete;

    bool onInit(EGLDisplay eglDisplay) noexcept override
    {
        m_cube = CubeMesh::create();
        if (!m_cube)
        {
            return false;
        }

        m_shader = PhongShader::create();
        if (!m_shader)
        {
            m_cube.reset();
            return false;
        }

        m_camera = Camera::create(m_width, m_height);
        if (m_camera)
        {
            // Create one texture per available camera buffer.
            for (const auto& request : m_camera->getRequests())
            {
                // We know that we are only using one stream and one buffer per
                // request. If we were using multiple streams at once, we
                // should iterate on the request BufferMap.
                auto [stream, buffer] = *request->buffers().begin();

                auto texture = CameraTexture::create(eglDisplay, stream->configuration(), *buffer);
                if (!texture)
                {
                    std::cerr << "Failed to create a camera texture" << std::endl;

                    m_textures.clear();
                    m_camera.reset();
                    m_shader.reset();
                    m_cube.reset();

                    return false;
                }

                // We are using the associated buffer cookie to store the
                // corresponding texture index in the internal vector. This way
                // it will be easy to fetch the right texture when a frame
                // buffer is ready to be drawn.
                m_textures.push_back(std::move(texture));
                buffer->setCookie(m_textures.size() - 1);
            }

            m_camera->startCapturing();
        }

        glClearColor(0.2f, 0.2f, 0.2f, 1.0f);
        glEnable(GL_CULL_FACE);
        glEnable(GL_DEPTH_TEST);

        m_shader->use();

        m_shader->setMaterialAmbient(glm::vec3(0.2f, 0.1f, 0.1f));
        m_shader->setMaterialDiffuse(glm::vec3(1.0f, 0.2f, 0.2f));
        m_shader->setMaterialSpecular(glm::vec3(1.0f, 1.0f, 1.0f));
        m_shader->setMaterialShininess(4.0f);

        m_shader->setCameraTextureMix(0.0f);
        m_shader->setCameraAspectRatio(static_cast<float>(m_width) / static_cast<float>(m_height));

        m_shader->setLightPosition(glm::vec3(1.5f, 1.5f, 2.0f));
        m_shader->setLightColor(glm::vec3(0.8f, 1.0f, 0.3f));

        auto cameraPosition = glm::vec3(0.0f, 0.0f, 2.0f);
        m_shader->setViewPosition(cameraPosition);
        m_shader->setViewMatrix(glm::lookAt(cameraPosition, glm::vec3(0.0f), glm::vec3(0.0f, 1.0f, 0.0f)));

        return true;
    }

    void onShut() noexcept override
    {
        m_textures.clear();
        m_currentCameraFrame.reset();
        m_camera.reset();
        m_shader.reset();
        m_cube.reset();
    }

    void onFrameBufferResized(int width, int height) noexcept override
    {
        glViewport(0, 0, width, height);

        float aspect = static_cast<float>(width) / static_cast<float>(height);
        m_shader->setProjectionMatrix(glm::perspective(glm::radians(60.0f), aspect, 0.1f, 100.0f));
    }

    void onRender(double time) noexcept override
    {
        if (m_camera)
        {
            // We are fetching the next camera produced frame that is ready to
            // be drawn. If there is no new frame available, we are just
            // keeping on drawing the same frame.
            auto cameraFrame = m_camera->getNextFrame();
            if (cameraFrame)
            {
                // We need to keep a reference to the current drawn frame in
                // order to not have the Camera class recycle the underlying
                // dma-buf while the GLES renderer is still using it for
                // drawing. This is the Camera::Frame destructor which ensures
                // proper synchronization. When reaching this point, the
                // previous m_currentCameraFrame has been fully drawn (the GLES
                // buffers swap has just occurred on the previous onRender
                // call), when the unique_ptr is replaced the previous
                // Camera::Frame is destroyed which triggers the recycling of
                // its FrameBuffer (for the next camera frame capture), while
                // the new frame is locked for drawing until it is itself
                // replaced.
                m_currentCameraFrame = std::move(cameraFrame);

                // We can directly fetch and bind the corresponding GLES
                // texture from the FrameBuffer cookie.
                auto textureIndex = m_currentCameraFrame->getCookie();
                m_textures[textureIndex]->bind();

                // The texture mix value is only used to reuse the same shader
                // without and with a camera frame. Now that we have a frame to
                // draw we can show it.
                m_shader->setCameraTextureMix(1.0f);
            }
        }

        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

        glm::mat4 modelMatrix =
            glm::rotate(glm::mat4(1.0f), 1.5f * static_cast<float>(time), glm::vec3(0.8f, 0.5f, 0.4f));
        m_shader->setModelMatrix(modelMatrix);

        m_cube->draw();
    }

  private:
    const unsigned int m_width;
    const unsigned int m_height;

    std::unique_ptr<CubeMesh> m_cube;
    std::unique_ptr<PhongShader> m_shader;
    std::shared_ptr<Camera> m_camera;

    std::unique_ptr<Camera::Frame> m_currentCameraFrame;
    std::vector<std::unique_ptr<CameraTexture>> m_textures;
};

int main()
{
    static constexpr unsigned int FRAME_BUFFER_WIDTH = 800;
    static constexpr unsigned int FRAME_BUFFER_HEIGHT = 600;

    auto appWindow = AppWindow::create(FRAME_BUFFER_WIDTH, FRAME_BUFFER_HEIGHT, "Have fun with Cam and Berry",
                                       std::make_unique<AppRenderer>(FRAME_BUFFER_WIDTH, FRAME_BUFFER_HEIGHT));
    if (appWindow)
    {
        appWindow->runRenderLoop();
    }
    else
    {
        std::cerr << "Failed to create the application window" << std::endl;
    }

    AppWindow::terminate();
    return 0;
}
