// SPDX-License-Identifier: MIT
// SPDX-FileCopyrightText: 2025 Loïc Le Page <llepage@igalia.com>

#include "AppWindow.h"

#define GLFW_EXPOSE_NATIVE_EGL
#include <GLFW/glfw3native.h>

std::unique_ptr<AppWindow> AppWindow::create(int width, int height, const char* title,
                                             std::unique_ptr<IRenderer> renderer) noexcept
{
    if (!renderer || !glfwInit())
    {
        return nullptr;
    }

    glfwWindowHint(GLFW_CONTEXT_CREATION_API, GLFW_EGL_CONTEXT_API);
    glfwWindowHint(GLFW_CLIENT_API, GLFW_OPENGL_ES_API);
    glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 3);

    std::unique_ptr<AppWindow> wnd(new AppWindow());
    wnd->m_renderer = std::move(renderer);
    wnd->m_window = glfwCreateWindow(std::max(width, 1), std::max(height, 1), title, NULL, NULL);
    if (!wnd->m_window)
    {
        return nullptr;
    }

    EGLDisplay eglDisplay = glfwGetEGLDisplay();
    if (!eglDisplay)
    {
        return nullptr;
    }

    glfwMakeContextCurrent(wnd->m_window);

    glfwSetWindowUserPointer(wnd->m_window, wnd.get());
    glfwSetFramebufferSizeCallback(
        wnd->m_window, +[](GLFWwindow* window, int width, int height) {
            auto appWindow = static_cast<AppWindow*>(glfwGetWindowUserPointer(window));
            if (appWindow)
            {
                appWindow->m_renderer->onFrameBufferResized(width, height);
            }
        });

    if (!wnd->m_renderer->onInit(eglDisplay))
    {
        return nullptr;
    }

    glfwGetFramebufferSize(wnd->m_window, &width, &height);
    wnd->m_renderer->onFrameBufferResized(width, height);

    return wnd;
}

AppWindow::~AppWindow()
{
    m_renderer->onShut();
    glfwDestroyWindow(m_window);
}

void AppWindow::runRenderLoop() const noexcept
{
    while (!glfwWindowShouldClose(m_window))
    {
        m_renderer->onRender(glfwGetTime());
        glfwSwapBuffers(m_window);
        glfwPollEvents();
    }
}
