// SPDX-License-Identifier: MIT
// SPDX-FileCopyrightText: 2025 Loïc Le Page <llepage@igalia.com>

#include "PhongShader.h"

static const char s_vtxShaderCode[] = R"EOS(
#version 300 es

layout(location = 0) in vec3 vtxPosition;
layout(location = 1) in vec3 vtxNormal;
layout(location = 2) in vec2 vtxTexCoords;

out vec3 position;
out vec3 normal;
out vec2 texCoords;

uniform mat4 model;
uniform mat4 view;
uniform mat4 projection;
uniform float cameraAspectRatio;

void main()
{
    float ratio = 1.0 / cameraAspectRatio;

    position = (model * vec4(vtxPosition, 1.0)).xyz;
    normal = mat3(transpose(inverse(model))) * vtxNormal;
    texCoords = vec2(vtxTexCoords.x * ratio + 0.5 * (1.0 - ratio), vtxTexCoords.y);
    gl_Position = projection * view * vec4(position, 1.0);
}
)EOS";

static const char s_fragShaderCode[] = R"EOS(
#version 300 es
#extension GL_OES_EGL_image_external : require

precision mediump float;

in vec3 position;
in vec3 normal;
in vec2 texCoords;

layout(location = 0) out vec4 fragColor;

struct Material {
    vec3 ambient;
    vec3 diffuse;
    vec3 specular;
    float shininess;
};
uniform Material material;

uniform samplerExternalOES cameraTexture;
uniform float cameraTextureMix;

struct Light {
    vec3 position;
    vec3 color;
};
uniform Light light;
uniform vec3 viewPosition;

void main()
{
    vec3 normal = normalize(normal);
    vec3 texColor = mix(vec3(1.0), texture2D(cameraTexture, texCoords).rgb, cameraTextureMix);

    vec3 ambient = light.color * material.ambient * texColor;

    vec3 lightDir = normalize(light.position - position);
    float diffCoeff = max(dot(normal, lightDir), 0.0);
    vec3 diffuse = diffCoeff * light.color * material.diffuse * texColor;

    vec3 viewDir = normalize(viewPosition - position);
    vec3 reflectDir = reflect(-lightDir, normal);
    float specCoeff = pow(max(dot(viewDir, reflectDir), 0.0), material.shininess);
    vec3 specular = specCoeff * light.color * material.specular;

    fragColor = vec4(ambient + diffuse + specular, 1.0);
}
)EOS";

std::unique_ptr<PhongShader> PhongShader::create() noexcept
{
    std::unique_ptr<PhongShader> shader(new PhongShader());
    if (!shader->initGL(s_vtxShaderCode, s_fragShaderCode))
    {
        return nullptr;
    }

    shader->m_modelMatrixLocation = glGetUniformLocation(shader->program(), "model");
    shader->m_viewMatrixLocation = glGetUniformLocation(shader->program(), "view");
    shader->m_projectionMatrixLocation = glGetUniformLocation(shader->program(), "projection");
    shader->m_cameraAspectRatioLocation = glGetUniformLocation(shader->program(), "cameraAspectRatio");

    shader->m_materialAmbientLocation = glGetUniformLocation(shader->program(), "material.ambient");
    shader->m_materialDiffuseLocation = glGetUniformLocation(shader->program(), "material.diffuse");
    shader->m_materialSpecularLocation = glGetUniformLocation(shader->program(), "material.specular");
    shader->m_materialShininessLocation = glGetUniformLocation(shader->program(), "material.shininess");

    glUniform1i(glGetUniformLocation(shader->program(), "cameraTexture"), 0);
    shader->m_cameraTextureMixLocation = glGetUniformLocation(shader->program(), "cameraTextureMix");

    shader->m_lightPositionLocation = glGetUniformLocation(shader->program(), "light.position");
    shader->m_lightColorLocation = glGetUniformLocation(shader->program(), "light.color");

    shader->m_viewPositionLocation = glGetUniformLocation(shader->program(), "viewPosition");

    return shader;
}
